/* hashtable.c -- Universal Hash Table with Linear Probing
 * 
 * Copyright (C) 2004 H Zero Seven
 *
 * Authors: Christian Volmering <risk@h07.org>
 * 
 * This file is part of the aexpl-ng distribution.
 * 
 * The aexpl-ng distribution is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.
 *
 * The aexpl-ng distribution is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the aexpl-ng distribution; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 * MA 02111-1307 USA
 */

#include "config.h"
#include "hashtable.h"

#include <stdlib.h>
#include <string.h>

#ifdef DMALLOC
#  include <dmalloc.h>
#endif

#define INT32 sizeof (int)
#define RESIZE_FACTOR 2
#define RESIZE_LEVEL 50

struct _hnode_t
{
  const void *ptr;
  const void *key;
  struct _hnode_t *next;
};

struct _htable_t
{
  unsigned int size;
  unsigned int entries;
  unsigned int base;
  struct _hnode_t **buckets;
  struct _hfunc_t func;
  unsigned char opt;
};

static unsigned int _hash_str (const void *key, int table_size);
static int _equal_str (const void *l_key, const void *r_key);
static void *_copy_str (const void *key);
static void _dispose_str (const void *key);
static unsigned int _hash_long (const void *key, int table_size);
static int _equal_long (const void *l_key, const void *r_key);
static void *_copy_long (const void *key);
static void _dispose_long (const void *key);
static inline struct _hnode_t *_lookup (struct _htable_t *table,
					const void *key, int hash);
static inline int _optimize (struct _htable_t *table);

struct _hfunc_t KEY_STR = {
  _hash_str,
  _equal_str,
  _copy_str,
  _dispose_str
};

struct _hfunc_t KEY_LONG = {
  _hash_long,
  _equal_long,
  _copy_long,
  _dispose_long
};

/* {{{ _hash_str */
static unsigned int
_hash_str (const void *key, int table_size)
{
  char *pos = (unsigned char *) key;
  int i, part = strlen (key) / INT32;
  unsigned long hash = 0;

  for (i = 0; i < part; pos += INT32, ++i)
    hash += (long long) *(int *) pos;

  for (; *pos; ++pos)
    hash += (long long) *pos;

  return (hash % table_size);
}

/* }}} */

/* {{{ _equal_str */
static int
_equal_str (const void *l_key, const void *r_key)
{
  return (!strcmp ((char *) l_key, (char *) r_key));
}

/* }}} */

/* {{{ _copy_str */
static void *
_copy_str (const void *key)
{
  size_t len = strlen ((char *) key) + 1;
  char *copy = (char *) calloc (sizeof (char), len);

  if (copy)
    strncpy (copy, (char *) key, len);

  return (copy);
}

/* }}} */

/* {{{ _dispose_str */
static void
_dispose_str (const void *key)
{
  free ((char *) key);
}

/* }}} */

/* {{{ _hash_long */
static unsigned int
_hash_long (const void *key, int table_size)
{
  return (*(long *) key % table_size);
}

/* }}} */

/* {{{ _equal_long */
static int
_equal_long (const void *l_key, const void *r_key)
{
  return (*(long *) l_key == *(long *) r_key);
}

/* }}} */

/* {{{ _copy_long */
static void *
_copy_long (const void *key)
{
  long *copy = (long *) malloc (sizeof (long));

  if (copy)
    *copy = *(long *) key;

  return (copy);
}

/* }}} */

/* {{{ _dispose_long */
static void
_dispose_long (const void *key)
{
  free ((long *) key);
}

/* }}} */

/* {{{ _lookup */
static inline struct _hnode_t *
_lookup (struct _htable_t *table, const void *key, int hash)
{
  struct _hnode_t *node;

  for (node = table->buckets[hash]; node; node = node->next)
    if (table->func.equal (node->key, key))
      break;

  return (node);
}

/* }}} */

/* {{{ _optimize */
static inline int
_optimize (struct _htable_t *table)
{
  int fill_level = table->entries ? table->entries * 100 / table->size : 0;

  if (fill_level > RESIZE_LEVEL)
    return (htable_resize (table, table->size * RESIZE_FACTOR));
  else if ((table->size > table->base) && (fill_level < 100 - RESIZE_LEVEL))
    return (htable_resize (table, table->size / RESIZE_FACTOR));

  return (1);
}

/* }}} */

/* {{{ new_htable */
struct _htable_t *
new_htable (int size, int opt, struct _hfunc_t func)
{
  struct _htable_t *table = NULL;

  if (size > 0)
    table = (struct _htable_t *) malloc (sizeof (struct _htable_t));

  if (table)
    {
      table->size = 2;
      table->entries = 0;
      table->opt = opt;
      table->func = func;

      while (table->size < size)
	table->size <<= 1;
      table->base = table->size;

      if (!(table->buckets =
	    (struct _hnode_t **) calloc (table->size,
					 sizeof (struct _hnode_t *))))
	{
	  free (table);
	  table = NULL;
	}
    }

  return (table);
}

/* }}} */

/* {{{ free_htable */
void
free_htable (struct _htable_t *table)
{
  htable_clear (table);
  free (table->buckets);
  free (table);
}

/* }}} */

/* {{{ htable_insert */
int
htable_insert (struct _htable_t *table, const void *key, const void *ptr)
{
  int hash = table->func.hash (key, table->size);
  struct _hnode_t *node;

  if ((node = _lookup (table, key, hash)))
    {
      node->ptr = ptr;
      return (1);
    }
  else if (!(node = (struct _hnode_t *) malloc (sizeof (struct _hnode_t))))
    return (0);
  else
    {
      if (table->opt & OPT_STORE)
	node->key = table->func.copy (key);
      else
	node->key = key;

      if (!node->key)
	return (0);

      node->ptr = ptr;
      node->next = table->buckets[hash];
      table->buckets[hash] = node;
      table->entries++;

      if (table->opt & OPT_OPTIMIZE)
	return (_optimize (table));

      return (1);
    }
}

/* }}} */

/* {{{ htable_delete */
int
htable_delete (struct _htable_t *table, const void *key)
{
  int hash = table->func.hash (key, table->size);
  struct _hnode_t *prev, *node = _lookup (table, key, hash);

  if (!node)
    return (0);

  if (table->buckets[hash] == node)
    table->buckets[hash] = node->next;
  else
    {
      for (prev = table->buckets[hash]; prev; prev = prev->next)
	if (prev->next == node)
	  {
	    prev->next = node->next;
	    break;
	  }
    }

  table->entries--;
  if (table->opt & OPT_STORE)
    table->func.dispose (node->key);
  free (node);

  if (table->opt & OPT_OPTIMIZE)
    return (_optimize (table));

  return (1);
}

/* }}} */

/* {{{ htable_lookup */
const void *
htable_lookup (struct _htable_t *table, const void *key)
{
  struct _hnode_t *node = _lookup (table, key,
				   table->func.hash (key, table->size));
  const void *ptr = node ? node->ptr : NULL;

  return (ptr);
}

/* }}} */

/* {{{ htable_contains */
int
htable_contains (struct _htable_t *table, const void *key)
{
  return (_lookup (table, key, table->func.hash (key, table->size)) ? 1 : 0);
}

/* }}} */

/* {{{ htable_size */
int
htable_size (struct _htable_t *table)
{
  return (table->size);
}

/* }}} */

/* {{{ htable_entries */
int
htable_entries (struct _htable_t *table)
{
  return (table->entries);
}

/* }}} */

/* {{{ htable_clear */
int
htable_clear (struct _htable_t *table)
{
  struct _hnode_t *node, *next;
  int i = 0;

  for (; i < table->size; i++)
    {
      node = table->buckets[i];
      while (node)
	{
	  next = node->next;
	  if (table->opt & OPT_STORE)
	    table->func.dispose (node->key);
	  free (node);
	  node = next;
	}
      table->buckets[i] = NULL;
    }

  table->entries = 0;

  if (table->opt & OPT_OPTIMIZE)
    return (htable_resize (table, table->base));

  return (1);
}

/* }}} */

/* {{{ htable_resize */
int
htable_resize (struct _htable_t *table, int size)
{
  struct _hnode_t **buckets, *node, *next;
  int i, hash, new_size = 2;

  while (new_size < size)
    new_size <<= 1;

  if (!(buckets =
	(struct _hnode_t **) calloc (new_size, sizeof (struct _hnode_t *))))
    return (0);

  if (table->entries)
    {
      for (i = 0; i < table->size; i++)
	{
	  node = table->buckets[i];
	  while (node)
	    {
	      next = node->next;
	      hash = table->func.hash (node->key, new_size);
	      node->next = buckets[hash];
	      buckets[hash] = node;
	      node = next;
	    }
	}
    }

  table->size = new_size;
  free (table->buckets);
  table->buckets = buckets;

  return (1);
}

/* }}} */

/*
 * Local variables:
 * tab-width: 8
 * c-basic-offset: 8
 * End:
 * vim600: noet nonu sw=8 ts=8 fdm=marker
 * vim<600: noet nonu sw=8 ts=8
 */
